@@ -15,7 +15,10 @@ use cargo_metadata::{Dependency, DependencyKind, Metadata, NodeDep, Package, Res
 use once_cell::sync::OnceCell;
 use petgraph::prelude::*;
 use semver::Version;
-use std::collections::{BTreeMap, HashMap, HashSet};
+use std::{
+    borrow::Cow,
+    collections::{BTreeMap, HashMap, HashSet},
+};
 use target_spec::TargetSpec;
 
 impl PackageGraph {
@@ -199,11 +202,9 @@ impl<'a> GraphBuildState<'a> {
                     )));
                 }
             };
-            let rel_path = pathdiff::diff_paths(dirname, self.workspace_root)
+            let rel_path = pathdiff::diff_utf8_paths(dirname, self.workspace_root)
                 .expect("workspace root is absolute");
-            let rel_path = Utf8PathBuf::from_path_buf(rel_path)
-                .expect("diff between two UTF-8 paths should produce a UTF-8 path");
-            PackageSourceImpl::Path(rel_path.into_boxed_path())
+            PackageSourceImpl::Path(convert_forward_slashes(rel_path).into_boxed_path())
         };
 
         let mut build_targets = BuildTargets::new(&package_id);
@@ -293,6 +294,7 @@ impl<'a> GraphBuildState<'a> {
                 metadata_table: package.metadata,
                 links: package.links.map(|s| s.into()),
                 publish: PackagePublishImpl::new(package.publish),
+                rust_version: package.rust_version,
                 features,
 
                 package_ix,
@@ -337,7 +339,7 @@ impl<'a> GraphBuildState<'a> {
                 id, manifest_path
             ))
         })?;
-        Ok(workspace_path.to_path_buf().into_boxed_path())
+        Ok(convert_forward_slashes(workspace_path).into_boxed_path())
     }
 
     fn finish(self) -> Graph<PackageId, PackageLinkImpl, Directed, PackageIx> {
@@ -746,3 +748,35 @@ impl PackagePublishImpl {
         }
     }
 }
+
+/// Replace backslashes in a relative path with forward slashes on Windows.
+fn convert_forward_slashes<'a>(rel_path: impl Into<Cow<'a, Utf8Path>>) -> Utf8PathBuf {
+    let rel_path = rel_path.into();
+    debug_assert!(
+        rel_path.is_relative(),
+        "path {} should be relative",
+        rel_path,
+    );
+
+    cfg_if::cfg_if! {
+        if #[cfg(windows)] {
+            rel_path.as_str().replace("\\", "/").into()
+        } else {
+            rel_path.into_owned()
+        }
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn test_convert_forward_slashes() {
+        let components = vec!["..", "..", "foo", "bar", "baz.txt"];
+        let path: Utf8PathBuf = components.into_iter().collect();
+        let path = convert_forward_slashes(path);
+        // This should have forward slashes, even on Windows.
+        assert_eq!(path.as_str(), "../../foo/bar/baz.txt");
+    }
+}
